/* lab11a.c */

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#define I_MAX_VALOR     254
#define SZ_FORMATO      "Formato: p <#posicoes> <#produtores> <#consumidores>.\n"
#define SZ_MAX_THREADS  "O maximo de threads permitido e' %d.\n"
#define SZ_POS_POSICOES "O numero de posicoes deve ser positivo.\n"
#define SZ_MAX_POSICOES "O maximo de posicoes permitido e' %d.\n"
#define SZ_ERRO_CRIA    "Erro na criacao do thread %d.\n"
#define SZ_NUM_THREADS  "Numero de threads %d.\n"
#define SZ_CONSUMINDO   "Consumidor %d esta' consumindo mensagem %d.\n"
#define SZ_JA_LIDAS     "Consumidor %d lendo mensagem %d de %d.\n"
#define SZ_PRODUZINDO   "Produtor %d esta' produzindo a mensagem %d.\n"
#define SZ_JA_PROD      "Produtor %d colocando mensagem %d de %d.\n"
#define SZ_CONS_ESPERA  "Consumidor %d esperando mensagem.\n"
#define SZ_PROD_ESPERA  "Produtor %d esperando espaco no buffer.\n"
#define SZ_ERRO_COND    "Erro na inicializacao de condicao %d.\n"


typedef void ( *Callback ) ( void* Informacao ) ;

typedef struct {
   Callback Funcao ;
   void*    Info   ;
} tpAviso ;

typedef struct {
   pthread_mutex_t MonitorBuffer                        ;
   pthread_cond_t  CondProdutor                         ;
   int             front                                ,
                   rear                                 ,
                   count                                ,
                   Mensagens[ I_MAX_VALOR ]             ,
                   Posicao[ I_MAX_VALOR ]               ,
                   JaLeram[ I_MAX_VALOR ]               ,
                   JaLidas[ I_MAX_VALOR ]               ,
                   Esperando[ I_MAX_VALOR ]             ;
   tpAviso         Avisos[ I_MAX_VALOR ][ I_MAX_VALOR ] ;

} tpBuffer ;


int             b                                           , /* posicoes */
                n                                           , /* produtores */
                m                                           ; /* consumidores */
tpBuffer        Buffer        = { PTHREAD_MUTEX_INITIALIZER ,
                                  PTHREAD_COND_INITIALIZER  ,
                                  0, 0, 0 }                 ;
pthread_mutex_t MonConsumidor = PTHREAD_MUTEX_INITIALIZER   ;


void* ProcessoConsumidor( void* MeuId ) ;

void* ProcessoProdutor( void* MeuId ) ;

int retira( int      iMeuId ,
            void*    Info   ,
            Callback Funcao  ) ;

void deposita( int iMensagem ,
               int iMeuId     ) ;

void Esperar( int iMeuId ) ;

void Avisar( void* Info ) ;


void* ProcessoConsumidor( void* MeuId ) {
   pthread_cond_t* pCondCons                  ;
   int             iMeuId    = *( int* )MeuId ,
                   iMensagem                  ;

   pCondCons = ( pthread_cond_t* )malloc( sizeof( pthread_cond_t ) ) ;
   if ( pthread_cond_init( pCondCons, NULL ) != 0 ) {
      printf( SZ_ERRO_COND, iMeuId ) ;
      exit( 1 ) ;
   } /* if */ 

   while ( 1 ) {
      iMensagem = retira( iMeuId, ( void* )pCondCons, Avisar ) ;

      while ( iMensagem == -1 ) {
         pthread_mutex_lock( &MonConsumidor ) ;
         printf( SZ_CONS_ESPERA, iMeuId ) ;
         fflush( stdout ) ;
         pthread_cond_wait( pCondCons, &MonConsumidor ) ;
         pthread_mutex_unlock( &MonConsumidor ) ;
         iMensagem = retira( iMeuId, ( void* )pCondCons, Avisar ) ;
      } /* while */
         
      printf( SZ_CONSUMINDO, iMeuId, iMensagem ) ;
      fflush( stdout ) ;
      Esperar( iMeuId ) ;
   } /* while */
} /* ProcessoConsumidor */


void* ProcessoProdutor( void* MeuId ) {
   int iMeuId    = *( int* )MeuId ,
       iMensagem                  ;

   srand( time( NULL ) ) ;

   while ( 1 ) {
      iMensagem = 1 + ( int )( ( ( double )I_MAX_VALOR * rand( ) )
                             / ( RAND_MAX + 1.0 ) ) ;
      printf( SZ_PRODUZINDO, iMeuId, iMensagem ) ;
      fflush( stdout ) ;
      deposita( iMensagem, iMeuId ) ;
      Esperar( iMeuId ) ;
   } /* while */
} /* ProcessoProdutor */


int retira( int      iMeuId ,
            void*    Info   ,
            Callback Funcao  ) {
   int iMensagem ;

   pthread_mutex_lock( &Buffer.MonitorBuffer ) ;

   if ( Buffer.count == 0
     || Buffer.JaLidas[ iMeuId ] == Buffer.count ) {

      Buffer.Avisos[ Buffer.Posicao[ iMeuId ] ]
             [ Buffer.Esperando[ Buffer.Posicao[ iMeuId ] ] ].Funcao = Funcao ;

      Buffer.Avisos[ Buffer.Posicao[ iMeuId ] ]
             [ Buffer.Esperando[ Buffer.Posicao[ iMeuId ] ] ].Info = Info ;

      Buffer.Esperando[ Buffer.Posicao[ iMeuId ] ]++ ;
      pthread_mutex_unlock( &Buffer.MonitorBuffer ) ;
      return -1 ;

   } /* if */

   iMensagem = Buffer.Mensagens[ Buffer.Posicao[ iMeuId ] ] ;
   Buffer.JaLeram[ Buffer.Posicao[ iMeuId ] ]++ ;
   Buffer.JaLidas[ iMeuId ]++ ;

   printf( SZ_JA_LIDAS, iMeuId, Buffer.JaLidas[ iMeuId ], Buffer.count ) ;
   fflush( stdout ) ;

   if ( Buffer.JaLeram[ Buffer.Posicao[ iMeuId ] ] == m ) {
      int i ;

      Buffer.JaLeram[ Buffer.Posicao[ iMeuId ] ] = 0 ;
      Buffer.front = ( Buffer.front + 1 ) % b ;
      Buffer.count-- ;
      pthread_cond_signal( &Buffer.CondProdutor ) ;
      
      for ( i = 0 ; i < m ; i++ ) {
         Buffer.JaLidas[ i ]-- ;
      } /* for */

   } /* if */    

   Buffer.Posicao[ iMeuId ] = ( Buffer.Posicao[ iMeuId ] + 1 ) % b ;
   pthread_mutex_unlock( &Buffer.MonitorBuffer ) ;

   return iMensagem ;
} /* retira */


void deposita( int iMensagem ,
               int iMeuId     ) {
   
   int i ;

   pthread_mutex_lock( &Buffer.MonitorBuffer ) ;

   while ( Buffer.count == b ) {
      printf( SZ_PROD_ESPERA, iMeuId ) ;
      fflush( stdout ) ;
      pthread_cond_wait( &Buffer.CondProdutor, &Buffer.MonitorBuffer ) ;
   } /* while */

   Buffer.Mensagens[ Buffer.rear ] = iMensagem ;

   for ( i = 0 ; i < Buffer.Esperando[ Buffer.rear ] ; i++ ) {
      Buffer.Avisos[ Buffer.rear ][ i ].Funcao( Buffer.Avisos[ Buffer.rear ]
                                                             [ i ].Info ) ;
   } /* for */
   Buffer.Esperando[ Buffer.rear ] = 0 ;

   Buffer.rear = ( Buffer.rear + 1 ) % b ;
   Buffer.count++ ;

   printf( SZ_JA_PROD, iMeuId, Buffer.count, b ) ;
   fflush( stdout ) ;

   pthread_mutex_unlock( &Buffer.MonitorBuffer ) ;
} /* deposita */


void Avisar( void* Info ) {
   pthread_mutex_lock( &MonConsumidor ) ;
   pthread_cond_signal( ( pthread_cond_t* )Info ) ;
   pthread_mutex_unlock( &MonConsumidor ) ;
} /* Avisar */


void Esperar( int iMeuId ) {
   unsigned long int i                                        ,
                     lTotal = ( unsigned long )10000000L
                            * ( unsigned long )( iMeuId + 1 ) ;

   for ( i = 0 ; i < lTotal ; i++ ) { }
} /* Esperar */


int main( int   argc    ,
          char* argv[ ]  ) {

   int       i                     ,
             iErro                 ;
   int*      pMeuId                ;
   pthread_t Thread[ I_MAX_VALOR ] ;

   if ( argc != 4 ) {
      printf( SZ_FORMATO ) ;
      return 0 ;
   } /* if */

   b = atoi( argv[ 1 ] ) ;
   n = atoi( argv[ 2 ] ) ;
   m = atoi( argv[ 3 ] ) ;

   if ( n + m > I_MAX_VALOR ) {
      printf( SZ_MAX_THREADS, I_MAX_VALOR ) ;
      return 0 ;
   } /* if */

   if ( b <= 0 ) {
      printf( SZ_POS_POSICOES ) ;
      return 0 ;
   } /* if */

   if ( b > I_MAX_VALOR ) {
      printf( SZ_MAX_POSICOES, I_MAX_VALOR ) ;
      return 0 ;
   } /* if */

   printf( SZ_NUM_THREADS, n + m ) ;
   
   for ( i = 0 ; i < b ; i++ ) {
      Buffer.Mensagens[ i ] = -1 ;
      Buffer.JaLeram[ i ] = 0 ;
      Buffer.Esperando[ i ] = 0 ; 
   } /* for */

   for ( i = 0 ; i < m ; i++ ) {
      Buffer.Posicao[ i ] = 0 ;
      Buffer.JaLidas[ i ] = 0 ;
   } /* for */

   for ( i = 0 ; i < n + m ; i++ ) {
      pMeuId = ( int* )malloc( sizeof( int ) ) ;
      *pMeuId = i ;

      if ( i < m ) {
         iErro = pthread_create( &Thread[ i ], NULL, ProcessoConsumidor,
                                                     ( void* )( pMeuId ) ) ;
      } else {
         iErro = pthread_create( &Thread[ i ], NULL, ProcessoProdutor,
                                                     ( void* )( pMeuId ) ) ;
      } /* if */ 

      if ( iErro != 0 ) {
         printf( SZ_ERRO_CRIA, i ) ;
         return 1 ;
      } /* if */
   } /* for */

   pthread_exit( NULL ) ;

   return 0 ;
}
